/**
  *  \file client/chartusr.cc
  *  \brief Starchart -- User Interface
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart.h"
#include "client/chartloc.h"
#include "client/chartmode.h"
#include "client/chartusr.h"
#include "client/cm-message.h"
#include "client/cm-standard.h"
#include "client/cm-marker.h"
#include "client/cscreen.h"
#include "client/dlg-visual.h"
#include "client/dlg-draw.h"
#include "client/history.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "game/lock.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "game/pref.h"
#include "game/formula.h"
#include "client/cm-distance.h"
#include "int/intglobal.h"
#include "client/dlg-comment.h"
#include "client/dialogs/gotoxydlg.h"
#include "client/dialogs/labeldlg.h"
#include "util/modinit.h"
#include "game/global.h"
#include "int/if/userif.h"
#include "int/process.h"
#include "client/dialogs/consoledlg.h"
#include "int/keymap.h"

static GChartLocation chart_location;
static WChartModeList chart_modes;

/********************************** Main *********************************/

static void
initStarchart()
{
    chart_location.init();
}

static void
initStarchartAfterLoad()
{
    chart_location.initAfterLoad();
}

void
setStarchartPosition(GPoint pt)
{
    chart_location.setLocation(pt);
}

void
setStarchartObject(GObject& obj)
{
    chart_location.setCurrentObject(obj);
}

GPoint
getStarchartPosition()
{
    return chart_location.getLocation().getNormalized();
}

static void
doChartLock(GPoint& pt, TLockMode lm, bool marked)
{
    GLockData locker(pt, marked);
    findItem(getDisplayedTurn().getCurrentUniverse(), locker, lm);
    pt = locker.getPoint();
}

static void
doChartLock(TLockMode lm, bool marked)
{
    GPoint pt = chart_location.getLocation().getNormalized();
    doChartLock(pt, lm, marked);
    chart_location.setLocation(pt.getNearestTo(chart_location.getLocation()));
}

static void
doGoTo()
{
    GPoint pt;
    if (doGoToXYDialog(pt))
        setStarchartPosition(pt);
}

namespace {
    class StarchartWidget : public UIWidget, public IntUserInterfaceBinding {
     public:
        StarchartWidget();
        void drawContent(GfxCanvas&);
        bool handleEvent(const UIEvent& e, bool second_pass);

        void drawChart(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo&);

        // IntUserInterfaceBinding:
        int  getProperty(IntUserInterfaceProperty p);
        void doGoToScreen(int screen, int obj);
        void doGoToChart(GPoint pt);
        void doMoveScanner(GPoint pt);
        void doPopupConsole();
        void doOverlayMessage(string_t msg);
        void doLeaveRace();
        void doSetChartView(string_t msg);
        IntKeymapRef getKeymap();

        enum {
            kNone,
            kGoToScreen,
            kConsole,
            kLeaveChart,
            kLeaveRace
        } script_action;
        WScreenId script_screen_id;

        using IntUserInterfaceBinding::getPreviousBinding;
    };
}

StarchartWidget::StarchartWidget()
    : UIWidget(0),
      script_action(kNone)
{ }

void
StarchartWidget::drawContent(GfxCanvas& can)
{
    drawChart(can);
    can.update(getExtent());
}

bool
StarchartWidget::handleEvent(const UIEvent& /*e*/, bool /*second_pass*/)
{
    /* FIXME? */
    return true;
}

void
StarchartWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.pref_size = info.min_size = GfxPoint(100,100); // FIXME?
}

void
StarchartWidget::drawChart(GfxCanvas& can)
{
    GChartViewport& viewport = chart_location.getViewport();

    can.drawBar(viewport.view, standard_colors[0],
                standard_colors[0], SOLID_FILL,
                GFX_ALPHA_OPAQUE);
    viewport.draw(can,
                  getChartOpts(false, viewport.mult, viewport.divi),
                  getDisplayedTurn().getCurrentUniverse(),
                  getDisplayedTurn().getChartBBox());
    chart_modes.drawCursor(can, chart_location);
    chart_modes.drawOverlays(can, chart_location);
}

int
StarchartWidget::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
        return 4;
     case iuiScreenRegistered:
        return 1;
     case iuiSimFlag:
        return 0;
     case iuiScanX:
     case iuiChartX:
        return chart_location.getLocation().x;
     case iuiScanY:
     case iuiChartY:
        return chart_location.getLocation().y;
    }
    return -1;
}

void
StarchartWidget::doGoToScreen(int screen, int obj)
{
    if (screen == 4) {
        // We're already here
    } else if (screen == 0) {
        script_action = kLeaveChart;
    } else {
        script_action    = kGoToScreen;
        script_screen_id = WScreenId(WScreenId::ScreenId(screen), obj);
    }
}

void
StarchartWidget::doGoToChart(GPoint pt)
{
    setStarchartPosition(pt);
}

void
StarchartWidget::doMoveScanner(GPoint pt)
{
    setStarchartPosition(pt);
}

void
StarchartWidget::doPopupConsole()
{
    script_action = kConsole;
}

void
StarchartWidget::doOverlayMessage(string_t msg)
{
    WMessageChartMode::showMessage(chart_modes, msg);
}

void
StarchartWidget::doLeaveRace()
{
    script_action = kLeaveRace;
    getPreviousBinding()->doLeaveRace();
}

void
StarchartWidget::doSetChartView(string_t msg)
{
    chart_location.setViewName(msg);
}

IntKeymapRef
StarchartWidget::getKeymap()
{
    if (IntKeymapRef view_keymap = chart_location.getKeymap())
        return view_keymap;
    else
        return IntKeymap::getKeymapByName("STARCHART");
}

static void
doColorizeAdjacent(GDrawingContainer& c, GPoint pt, uint8 color)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getNormalized();
    while (i != e) {
        if (i->isVisible() && i->getType() == LineDrawing && i->getColor() != color) {
            if (i->getPos().getNormalized() == pt) {
                // Starts at given position
                i->setColor(color);
                pt = i->getPos2().getNormalized();
                i = c.begin();
            } else if (i->getPos2().getNormalized() == pt) {
                // Ends at given position
                i->setColor(color);
                pt = i->getPos().getNormalized();
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

static void
doColorChange(GChartLocation& loc)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getNormalized());
    if (i == c.end())
        return;

    // Change color
    WColorSelector csel;
    csel.setColor(GDrawing::getUserColorFromColor(i->getColor()));
    int n = csel.doStandardDialog(_("Drawing Color"), (i->getType() == LineDrawing));
    if (n == csel.rCancel)
        return;

    uint8 new_color = GDrawing::getColorFromUserColor(csel.getColor());
    i->setColor(new_color);
    if (n == csel.rAdjacent) {
        doColorizeAdjacent(c, i->getPos(), new_color);
        doColorizeAdjacent(c, i->getPos2(), new_color);
    }

    loc.markChanged();
}

static void
doMoveMarker(GChartLocation& loc)
{
    // Not possible if drawings not visible
    GChartViewport& viewport = loc.getViewport();
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find drawing here
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.begin();
    while (i != c.end() && (i->getPos().getNormalized() != loc.getLocation().getNormalized() || i->getType() != MarkerDrawing))
        ++i;

    if (i != c.end())
        WMarkerChartMode::moveMarker(chart_modes, i);
}

static void
doDraw(GChartLocation& loc)
{
    // Ask user what to do
    WNewDrawingDialog ndd;
    ndd.setColor(9);
    if (!ndd.runDialog())
        return;

    // Draw it
    GDrawing d(loc.getLocation().getNormalized(), ndd.getType());
    d.setColor(GDrawing::getColorFromUserColor(ndd.getColor()));
    d.setTag(ndd.getTag());
    switch (ndd.getType()) {
     case LineDrawing:
     case RectangleDrawing:
        d.setPos2(d.getPos());
        WMarkerChartMode::drawLineRectangle(chart_modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
     case MarkerDrawing: {
        WMarkerTypeSelector mts;
        WColorSelector cs;
        if (mts.doStandardDialog(_("New marker"))) {
            d.setMarkerKind(mts.getMarkerType());
            cs.setColor(GDrawing::getUserColorFromColor(d.getColor()));
            if (cs.doStandardDialog(_("New marker"), false) == cs.rOK) {
                d.setColor(GDrawing::getColorFromUserColor(cs.getColor()));
                getDisplayedTurn().getCurrentUniverse().getDrawings().add(d);
                loc.markChanged();
            }
        }
        delete chart_modes.getModeByKind(ChartMode_PrimaryMode);
        break;
     }
     case CircleDrawing:
        d.setCircleRadius(10);
        WMarkerChartMode::drawCircle(chart_modes, getDisplayedTurn().getCurrentUniverse().getDrawings().add(d));
        break;
    }
}

static void
doDistance(GChartLocation& loc, WChartModeList& modes)
{
    modes.pushMode(*new WDistanceChartMode(loc.getLocation(), loc.getCurrentShipId()));
    if (loc.getCurrentShipId() != 0) {
        GShip& sh = getDisplayedTurn().getCurrentUniverse().getShip(loc.getCurrentShipId());
        if (sh.getKind() == GShip::CurrentShip) {
            setStarchartPosition(sh.getWaypoint());
        }
    }
}

static void
doDelete(GChartLocation& loc)
{
    GChartViewport& viewport = loc.getViewport();

    // Not possible if drawings not visible
    if ((getChartOpts(false, viewport.mult, viewport.divi).show & GChartOptions::co_Drawings) == 0)
        return;

    // Find nearest visible drawing
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator i = c.findNearestVisibleDrawing(loc.getLocation().getNormalized());
    if (i == c.end())
        return;

    WMarkerChartMode::deleteDrawing(chart_modes, i);
}

/** Page through objects.
    \param loc Chart location object to operate on
    \param delta Browse direction (+1/-1)
    \param marked_only true to accept only marked objects
    \param any true to accept any race's object. Default is to accept only played objects
           when starting at a played object.

    FIXME: PCC 1.x behaves a little different. With any=false, PCC 1.x would only accept
    played objects no matter where we start, but it would "degrade" to accept non-played
    objects when there are no played ones. */
static void
doPage(GChartLocation& loc, int delta, bool marked_only, bool any)
{
    GObjectType* type;

    /* Fetch current object, and associated type object */
    GObjectId id = loc.getCurrentObject();
    if (!id.isValid() || !id.isGameObject())
        return;

    if (GShip* sh = dynamic_cast<GShip*>(&id.getObject())) {
        // Iterate through all ships or player ships
        if (sh->isPlayed() && !any)
            type = &global_ship_type;
        else
            type = &global_anyship_type;
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&id.getObject())) {
        // Iterate through all planets or player planets
        if (pl->isPlayed() && !any)
            type = &global_planet_type;
        else
            type = &global_anyplanet_type;
    } else {
        // Iterate through whatever it is
        type = &id.getType();
    }

    /* Get next object */
    int next;
    if (delta > 0)
        next = type->getNextWrap(id.getIndex(), marked_only);
    else
        next = type->getPreviousWrap(id.getIndex(), marked_only);

    /* Valid? */
    if (type->isExistant(next))
        setStarchartObject(type->getObject(next));
}

static void
tryEditMarkerComment(GChartLocation& loc)
{
    /* Find a marker */
    GPoint pt = loc.getLocation().getNormalized();
    GDrawingContainer& drawings = getDisplayedTurn().getCurrentUniverse().getDrawings();
    GDrawingContainer::iterator p = drawings.begin();
    while (p != drawings.end()
           && !(p->isVisible()
                && p->getPos().getNormalized() == pt
                && p->getType() == MarkerDrawing))
        ++p;

    /* Edit it */
    if (p != drawings.end()) {
        editMarkerComment(*p);
        loc.markChanged();
    }
}

void
doStarchart()
{
    chart_location.initEntry(GfxRect(*ui_screen));
    chart_location.setCenter(chart_location.getLocation());

    StarchartWidget chart_w;
    chart_w.setExtent(chart_location.getViewport().view);
    chart_modes.deleteAllModes();
    chart_modes.pushMode(*new WStandardChartMode());

    int pending_dx = 0, pending_dy = 0;

    int speed = 0;
    int num_ticks_elapsed = 0;

    // FIXME: handle turn change
    SignalConnection conn_univ_changed(getDisplayedTurn().getCurrentUniverse()
                                         .sig_universe_changed.add(&chart_location, &GChartLocation::markChanged));

    ui_root->add(chart_w);
    clearSkippedTicks();
    while (1) {
        ui_root->performDeferredRedraws();

        /* If we have any scripts, run them */
        if (chart_w.script_action != StarchartWidget::kNone) {
            runRunnableProcesses();
            killTerminatedProcesses();
        }

        /* Recenter map */
        if (chart_location.getCenter() != chart_location.getLocation()) {
            if (speed == 0
                && (distanceSquared(chart_location.getLocation(), chart_location.getCenter())
                    <= square(getUserPreferences().ChartAnimThreshold())))
            {
                /* Do single small movement immediately */
                chart_location.setCenter(chart_location.getLocation());
                speed = 0;
            } else {
                /* Big movement only after tick event */
                while (num_ticks_elapsed > 0) {
                    --num_ticks_elapsed;
                    GPoint a = chart_location.getLocation();
                    GPoint b = chart_location.getCenter().getNearestTo(a);
                    double dist = std::sqrt(double(distanceSquared(a, b)));

                    if (dist <= speed) {
                        // We're very close, so go directly
                        chart_location.setCenter(chart_location.getLocation());
                        speed = 0;
                        num_ticks_elapsed = 0;
                    } else {
                        // We're a little farther. Compute speed.
                        if (dist < speed*speed) {
                            if (speed > 1)
                                --speed;
                        } else {
                            ++speed;
                        }

                        // Compute new location.
                        int distx = int((a.x - b.x) * speed / dist);
                        int disty = int((a.y - b.y) * speed / dist);

                        // Ensure we make progress.
                        if (distx == 0 && a.x != b.x)
                            distx = (a.x < b.x ? -1 : +1);
                        if (disty == 0 && a.y != b.y)
                            disty = (a.y < b.y ? -1 : +1);

                        // Move.
                        chart_location.setCenter(GPoint(b.x + distx, b.y + disty));
                    }
                }
            }
        }

        /* Process location and mode changes. Those cause a full redraw. */
        if (chart_location.isChanged() || chart_modes.isChanged()) {
            /* Let mode handle updates */
            chart_modes.onMove(chart_location);

            /* Redraw map */
            ui_screen->lock();
            chart_w.drawChart(*ui_screen);
            ui_screen->unlock();
            ui_screen->update(chart_w.getExtent());

            /* Clear change flags */
            chart_location.clearChanged();
            chart_modes.clearChanged();
            chart_modes.markClean();
        }

        /* Handle chart mode dirty regions. Those cause a partial redraw. */
        if (chart_modes.isDirty()) {
            GfxClipRectFilter filter(chart_modes.getDirtyRect(), *ui_screen);
            filter.lock();
            chart_w.drawChart(filter);
            filter.unlock();
            filter.update(chart_modes.getDirtyRect());
            chart_modes.markClean();
        }

        /* Do we have any script action to perform? If we have, do it,
           otherwise, fetch an event and process that. */
        UIEvent e;
        e.type = ev_Idle;
        switch (chart_w.script_action) {
         case StarchartWidget::kLeaveRace:
         case StarchartWidget::kLeaveChart:
            goto out;
         case StarchartWidget::kConsole:
            chart_w.script_action = StarchartWidget::kNone;
            doConsole();
            break;
         case StarchartWidget::kGoToScreen:
            chart_w.script_action = StarchartWidget::kNone;
            if (chart_w.script_screen_id.getScreenId() == WScreenId::Starchart) {
                chart_location.setLocation(chart_w.script_screen_id.getPos());
            } else if (chart_w.script_screen_id.isValid()) {
                chart_w.getPreviousBinding()->doGoToScreen(chart_w.script_screen_id.getScreenId(), chart_w.script_screen_id.getId());
                goto out;
            }
            break;
         case StarchartWidget::kNone:
            pushHistoryScreenId(WScreenId(chart_location.getLocation()));
            getEvent(e, false, true);
            break;
        }

        /* Drive timer */
        if (e.type == ev_Tick)
            num_ticks_elapsed = e.tick.elapsed;
        else
            num_ticks_elapsed = 0;

        if (chart_modes.handleEvent(e, chart_location)) {
            /* Handled by mode */
        } else if (e.type == ev_Mouse) {
            /* Mouse movement. When mouse movement does not immediately
               translate into screen movement, we accumulate mouse movement
               in pending_d[xy] instead of discarding it. This gives much
               smoother movement in large zoom levels. */
            pending_dx += e.mouse.x;
            pending_dy += e.mouse.y;
            GPoint new_pt = chart_location.getLocation();
            new_pt.x += chart_location.getViewport().unscale(pending_dx);
            new_pt.y -= chart_location.getViewport().unscale(pending_dy);

            if (e.mouse.b) {
                // mouse button held down: lock onto object
                if (e.mouse.b & mb_Left)
                    doChartLock(new_pt, lm_Left, e.mouse.b & ss_Ctrl);
                else if (e.mouse.b & mb_Right)
                    doChartLock(new_pt, lm_Right, e.mouse.b & ss_Ctrl);
            } else if (chart_location.getObjectList().getNumObjects() != 0) {
                // sticky mouse
                if (square(pending_dx) + square(pending_dy) <= square(getUserPreferences().ChartMouseStickiness())) {
                    new_pt = chart_location.getLocation();
                }
            } else {
                // regular movement
            }

            if (new_pt != chart_location.getLocation()) {
                chart_location.setLocation(new_pt);
                pending_dx = pending_dy = 0;
            }
        } else if (e.type == ev_Key) {
            /* Keys */
            switch (e.key.code) {
             case SDLK_RETURN:
                doChartLock(lm_Right, false);
                break;
             case SDLK_SPACE:
                doChartLock(lm_Left, false);
                break;
             case ss_Ctrl + SDLK_RETURN:
                doChartLock(lm_Right, true);
                break;
             case ss_Ctrl + SDLK_SPACE:
                doChartLock(lm_Left, true);
                break;
             case '-':
             case vk_WheelDown:
                chart_location.doZoom(false);
                break;
             case '+':
             case vk_WheelUp:
                chart_location.doZoom(true);
                break;
             case SDLK_F9:
             case SDLK_F9 + ss_Alt:
                // ship and planet comments are handled through the script interface
                tryEditMarkerComment(chart_location);
                break;
             case SDLK_TAB:
                chart_location.doScroll(+1, false);
                break;
             case ss_Shift + SDLK_TAB:
                chart_location.doScroll(-1, false);
                break;
             case ss_Ctrl + SDLK_TAB:
                chart_location.doScroll(+1, true);
                break;
             case ss_Shift + ss_Ctrl + SDLK_TAB:
                chart_location.doScroll(-1, true);
                break;
             case SDLK_PAGEDOWN:
                doPage(chart_location, +1, false, false);
                break;
             case SDLK_PAGEDOWN + ss_Ctrl:
                doPage(chart_location, +1, true, false);
                break;
             case SDLK_PAGEUP:
                doPage(chart_location, -1, false, false);
                break;
             case SDLK_PAGEUP + ss_Ctrl:
                doPage(chart_location, -1, true, false);
                break;
             case SDLK_PAGEDOWN + ss_Shift:
                doPage(chart_location, +1, false, true);
                break;
             case SDLK_PAGEDOWN + ss_Ctrl + ss_Shift:
                doPage(chart_location, +1, true, true);
                break;
             case SDLK_PAGEUP + ss_Shift:
                doPage(chart_location, -1, false, true);
                break;
             case SDLK_PAGEUP + ss_Ctrl + ss_Shift:
                doPage(chart_location, -1, true, true);
                break;
             case 'm':
                if (int mf = findNearestMinefield(chart_location.getLocation())) {
                    global_minefield_type.setCurrent(mf);
                    doMinefieldScreen();
                } else {
                    WMessageChartMode::showMessage(chart_modes, _("No Minefields"));
                }
                break;
             case 'u':
                if (int ufo = findNearestUfo(chart_location.getLocation())) {
                    global_ufo_type.setCurrent(ufo);
                    doUfoScreen();
                } else {
                    WMessageChartMode::showMessage(chart_modes, _("No Ufos"));
                }
                break;
             case 'c':
                doColorChange(chart_location);
                break;
             case 'v':
                doMoveMarker(chart_location);
                break;
             case 'p':
                doDraw(chart_location);
                break;
             case 'd':
                doDistance(chart_location, chart_modes);
                break;
             case 'g':
                doGoTo();
                break;
             case SDLK_DELETE:
                doDelete(chart_location);
                break;
             case ss_Ctrl + ss_Alt + 's':
                doScreenshot();
                break;
             case ss_Alt + 'm':
                chart_location.toggleOptions(GChartOptions::co_Mine);
                break;
             case ss_Alt + 'a':
                chart_location.toggleOptions(GChartOptions::co_ShipDots);
                break;
             case ss_Alt + 'd':
                chart_location.toggleOptions(GChartOptions::co_Labels);
                break;
             case ss_Alt + 'i':
                chart_location.toggleOptions(GChartOptions::co_Ion);
                break;
             case ss_Alt + 'v':
                chart_location.toggleOptions(GChartOptions::co_Trails);
                break;
             case ss_Alt + 't':
                chart_location.toggleOptions(GChartOptions::co_Selection);
                break;
             case ss_Alt + 's':
                chart_location.toggleOptions(GChartOptions::co_Sectors);
                break;
             case ss_Alt + 'b':
                chart_location.toggleOptions(GChartOptions::co_Borders);
                break;
             case ss_Alt + 'u':
                chart_location.toggleOptions(GChartOptions::co_Ufo);
                break;
             case ss_Alt + 'p':
                chart_location.toggleOptions(GChartOptions::co_Drawings);
                break;
             case ss_Alt + 'w':
                chart_location.toggleOptions(GChartOptions::co_WarpWells);
                break;
             case ss_Alt + 'l':
                editLabelExpressions();
                chart_location.markChanged();     // Force redraw
                break;
            }
        }
    }
 out:
    ui_root->close(chart_w);
    chart_modes.deleteAllModes();
}

const ModuleInitializer mod_chartusr[] = {
    { OnEnterNoGame | OnEnterDirectory | OnEnterGame, initStarchart },
    { OnEnteredGame, initStarchartAfterLoad },
    { 0, 0 },
};
