/**
  *  \file client/dlg-component.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-component.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/prefix.h"
#include "client/dlg-tech.h"       // WCostDisplay
#include "game/actions/baseparts.h"
#include "game/actions/basetech.h"
#include "ui/widgets/multilinestatic.h"

namespace {
    class WComponentCountDisplay : public UIWidget {
        GStarbaseBuildPartsAction& action;
        TTechLevel kind;
        int slot;
        SignalConnection conn;
     public:
        WComponentCountDisplay(GStarbaseBuildPartsAction& action, TTechLevel kind, int slot);
        ~WComponentCountDisplay();
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
    };

    class WBuildComponentWindow : public UIWindow {
        ObjectHolder holder;
        GStarbaseBuildPartsAction& action;
        GStarbaseBuildTransaction& ta;
        TTechLevel kind;
        int slot;
        SignalConnection conn;
        WCostDisplay& cost_dpy;
     public:
        WBuildComponentWindow(GStarbaseBuildPartsAction& action, GStarbaseBuildTransaction& ta, TTechLevel kind, int slot, const GComponent& component);
        ~WBuildComponentWindow();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void onChange();
    };
}

WComponentCountDisplay::WComponentCountDisplay(GStarbaseBuildPartsAction& action, TTechLevel kind, int slot)
    : action(action), kind(kind), slot(slot),
      conn(observe(action.sig_changed))
{ }

WComponentCountDisplay::~WComponentCountDisplay()
{ }

void
WComponentCountDisplay::drawContent(GfxCanvas& can)
{
    const GfxRect& pos = getExtent();
    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(false);
    s.setColor(UIColor::tc_Static);
    s.outText(can, FONT_TITLE, pos.x, pos.y, _("Current Amount:"));
    s.outText(can, FONT_TITLE, pos.x, pos.y + font_heights[FONT_TITLE], _("New Amount:"));
    s.setSolid(true);
    s.getParameters().just_x = 2;
    s.outTextF(can, FONT_TITLE, pos.x + pos.w, pos.y, 5*font_ems[FONT_TITLE],
               numToString(action.getExistingParts(kind, slot)));
    s.outTextF(can, FONT_TITLE, pos.x + pos.w, pos.y + font_heights[FONT_TITLE], 5*font_ems[FONT_TITLE],
               numToString(action.getTotalParts(kind, slot)));

    can.update(pos);
}

void
WComponentCountDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_TITLE].scaledBy(15, 2);
    info.flags = 0;
}

WBuildComponentWindow::WBuildComponentWindow(GStarbaseBuildPartsAction& action,
                                             GStarbaseBuildTransaction& ta,
                                             TTechLevel kind, int slot,
                                             const GComponent& component)
    : UIWindow(_("Build Components"), BLUE_WINDOW, COLOR_GRAY),
      action(action), ta(ta), kind(kind), slot(slot),
      conn(ta.sig_cost_changed.add(this, &WBuildComponentWindow::onChange)),
      cost_dpy(holder.add(new WCostDisplay(ta, GCargoTypeSet(el_Money) | el_Tritanium | el_Duranium | el_Molybdenum)))
{
    setLayout(UIVBoxLayout::instance5);

    /* Intro text */
    static const char*const messages[] = {
        N_("Use this window to build %s engines and put them into starbase storage for later use."),
        N_("Use this window to build %s hulls and put them into starbase storage for later use."),
        N_("Use this window to build %s beams and put them into starbase storage for later use."),
        N_("Use this window to build %s launchers and put them into starbase storage for later use."),
    };
    add(holder.add(new UIMultilineStatic(FONT_NORMAL,
                                         format(_(messages[kind]), component.getName()),
                                         20*font_ems[FONT_TITLE])));

    /* Controls */
    UIGroup& h0 = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& v0 = holder.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& h2 = holder.add(new UIGroup(UIHBoxLayout::instance5));
    add(h0);
    h0.add(holder.add(new WComponentCountDisplay(action, kind, slot)));
    h0.add(v0);

    v0.add(h2);
    v0.add(holder.add(new UISpacer()));

    h2.add(holder.add(new UIButton(cm_Next, "+", '+')).addCommand());
    h2.add(holder.add(new UIButton(cm_Prev, "-", '-')).addCommand());
    h2.add(holder.add(new UISpacer()));

    /* Cost */
    add(cost_dpy);

    /* OK/Cancel */
    UIGroup& h1 = holder.add(new UIGroup(UIHBoxLayout::instance5));
    add(h1);
    h1.add(holder.add(new UISpacer()));
    h1.add(holder.add(new UIButton(1001, _("OK"), SDLK_RETURN)).addStop());
    h1.add(holder.add(new UIButton(1000, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(holder.add(new UIPrefixArg()));

    pack();
}
               
WBuildComponentWindow::~WBuildComponentWindow()
{ }

bool
WBuildComponentWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_Next:
        action.addParts(kind, slot, cmd.arg ? cmd.arg : 1, false);
        return true;
     case cm_Prev:
        action.addParts(kind, slot, cmd.arg ? -cmd.arg : -1, false);
        return true;
     default:
        return UIWindow::handleCommand(cmd, sender);
    }
}

void
WBuildComponentWindow::onChange()
{
    cost_dpy.drawWidget();
    setStateOf(1001, st_Disabled, !ta.isValidCost());
}

void
doBuildComponent(GGameTurn& turn, int pid,
                 TTechLevel kind, int slot,
                 const GComponent& component)
{
    /* First, ask user for tech upgrade */
    GPlanet& p = turn.getCurrentUniverse().getPlanet(pid);
    int have_tech = p.getBaseTech(kind);
    int need_tech = component.getTech();
    if (need_tech > have_tech) {
        static const char*const messages[] = {
            N_("To build this engine, you need tech %d."),
            N_("To build this hull, you need tech %d."),
            N_("To build this beam, you need tech %d."),
            N_("To build this torpedo launcher, you need tech %d.")
        };

        /* Try tech upgrade */
        GPlanetTransfer pay(p);
        GStarbaseTechUpgradeAction action(p);
        GStarbaseBuildTransaction ta(action, pay);

        if (!action.setTech(kind, need_tech)) {
            messageBox(string_t(format(_(messages[kind]), need_tech))
                       + string_t(format(_(" You only have tech %d. As an unregistered player,"
                                           " you cannot upgrade to the required level."),
                                         have_tech)),
                       _("Build Components"));
            ta.doCancel();
            pay.doCancel();
            return;
        } else if (!ta.isValidCost()) {
            messageBox(string_t(format(_(messages[kind]), need_tech))
                       + string_t(format(_(" You do not have the %d megacredits required"
                                           " to upgrade to the required level."),
                                         action.getCost().get(el_Money))),
                       _("Build Components"));
            ta.doCancel();
            pay.doCancel();
            return;
        } else {
            /* gcc-3.3 misparses the code if I write the UIMessageBox
               initialisation as a single expression, hence here I spell
               it out. */
            string_t msg_text = format(_(messages[kind]), need_tech);
            msg_text += format(_(" Do you want to upgrade for %d mc?"),
                               action.getCost().get(el_Money));
            UIMessageBox msg(msg_text, _("Build Components"));
            if (msg.doYesNoDialog()) {
                ta.doCommit();
                pay.doCommit();
                turn.getCurrentUniverse().doScreenUpdates();
            } else {
                ta.doCancel();
                pay.doCancel();
                return;
            }
        }
    }

    /* At this point, we have appropriate tech and can start the build */
    GPlanetTransfer pay(p);
    GStarbaseBuildPartsAction action(p);
    GStarbaseBuildTransaction ta(action, pay);
    WBuildComponentWindow window(action, ta, kind, slot, component);
    action.addUndoInformation(turn.getPreviousUniverse().getPlanet(pid));

    ui_root->centerWidget(window);
    ui_root->add(window);
    int rv = ui_root->run();
    ui_root->close(window);

    if (rv == 1001 && !ta.getError()) {
        ta.doCommit();
        pay.doCommit();
        turn.getCurrentUniverse().doScreenUpdates();
    } else {
        ta.doCancel();
        pay.doCancel();
    }
}
