/**
  *  \file client/widgets/searchobj.cc
  *  \brief Search Query Object Selector
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/searchobj.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "client/search.h"

/** Constructor.
    \param id Widget Id */
WSearchObjectSelector::WSearchObjectSelector(int id)
    : UIListbox(id, 5, font_ems[FONT_NORMAL]*10, 5),
      objs(TUserSearchQuery::AllObjects)
{
    sig_dblclick.add(this, &WSearchObjectSelector::onToggle);
}

/** Destructor. */
WSearchObjectSelector::~WSearchObjectSelector()
{ }

/** Set object selection (bitmask).
    \param objs Selected objects, see TUserSearchQuery */
void
WSearchObjectSelector::setObjects(uint16_t objs)
{
    if (this->objs != objs) {
        this->objs = objs;
        drawWidget();
    }
}

/** Get object selection (bitmask). */
uint16_t
WSearchObjectSelector::getObjects() const
{
    return this->objs;
}

void
WSearchObjectSelector::drawPart(GfxCanvas& can, int from, int to)
{
    // xref TUserSearchQuery::SearchObject
    static const char* const names[] = {
        N_("Starships"),
        N_("Planets"),
        N_("Starbases"),
        N_("Ufos"),
        N_("Others"),
    };

    for (int i = from; i < to; ++i) {
        int line = getTop() + i;
        int y = getExtent().y + i*font_heights[FONT_NORMAL];
        int mid = 2*font_ems[FONT_NORMAL];

        /* Set parameters */
        const UIColorSet& scheme = (line != getCurrent()
                                    ? GRAY_COLOR_SET
                                    : getFocusState() != NoFocus
                                    ? BLACK_COLOR_SET
                                    : BRIGHT_GRAY_COLOR_SET);

        GfxParameters parm;
        parm.bgcolor = standard_colors[scheme[UIColor::tc_Background]];
        parm.color   = standard_colors[scheme[UIColor::tc_Static]];
        parm.font    = fonts[FONT_NORMAL];
        parm.just_x  = 0;
        parm.just_y  = 0;

        if (line < 0 || line >= int(countof(names))) {
            /* Error or widget too tall */
            outTextF(can, parm, getExtent().x,     y, getExtent().w,     string_t());
        } else {
            /* Draw it */
            // FIXME: use proper unicode character here
            outTextF(can, parm, getExtent().x,     y, mid,               (objs & (1 << line)) ? " \xD7" : " \xB7");
            outTextF(can, parm, getExtent().x+mid, y, getExtent().w-mid, _(names[line]));
        }
    }
}

bool
WSearchObjectSelector::handleEvent(const UIEvent& event, bool second_pass)
{
    if (hasState(st_Focused) && event.type == ev_Key && event.key.code == ' ') {
        sig_dblclick.raise(getId());
        return true;
    }
    return UIListbox::handleEvent(event, second_pass);
}

void
WSearchObjectSelector::onToggle()
{
    this->objs ^= 1 << getCurrent();
    drawWidget();
}
